/* 
 * NOTE: this will only work on one file at a time
 * TO make this work with more than one, we would have to keep state for each open file
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include "util/util_all.h"
#include "net/net_all.h"

#include "asf.h"
#include "misc.h"
#include "guids.h"

// Options supported:
//#define DEBUG
//#define QUIET
//#define SHOW_CONTENT_TYPE

#if defined(DEBUG) && defined(QUIET)
#error "DEBUG and QUIET are mutually exclusive"
#endif

unsigned char s_header[BUFFERSIZE], s_data[BUFFERSIZE];
unsigned long s_header_size = 0, s_file_header_size = 0;

// TODO: adjust this method to support more than one file (i.e., be asynchronous)
struct asf_header *s_asfh = NULL;
struct asf_file_header *s_asffh = NULL;
struct asf_data_subchunk_header *s_asfdsh = NULL;

////////////////////////////////////////////////////////////////
// ASF file creation/closing
 
/*
 * asf_create_file
 */
int asf_create_file(char *filename)
{
	int fd;
 
	assert(filename);
	if (!filename) return ERROR;
	
 	fd = open(filename, O_CREAT | O_TRUNC | O_RDWR, 0644);
	if (fd == ERROR)
	{
		fprintf(stderr, "Error creating/opening %s: %s\n", filename, strerror(errno));
		return ERROR;
	}

	return fd; 
}

/*
 * asf_append_file
 */
int asf_append_file(char *filename)
{
	int fd;
 
	assert(filename);
	if (!filename) return ERROR;

	fd = open(filename, O_RDWR | O_APPEND);
	if (fd == ERROR)
	{
		fprintf(stderr, "Error creating/opening %s: %s\n", filename, strerror(errno));
		return ERROR;
	}

	return fd;
}

/*
 * asf_reset_file
 */
int asf_reset_file(int fd)
{
	assert(fd >= 0);
	if (fd < 0) return ERROR;

	if (lseek(fd, 0, SEEK_SET) == ERROR)
	{
  	fprintf(stderr, "Error with lseek on fd %d: %s\n", fd, strerror(errno));
		return ERROR;
	}
 
	return 0;
}
 
/*
 * asf_close_file
 */
void asf_close_file(int fd)
{
	assert(fd >= 0);
	if (fd < 0) return;

  close(fd);
}

////////////////////////////////////////////////////////////////
// ASF file reading/writing
 
/*
 * asf_read_file
 */
int asf_read_file(char *filename)
{
  int numbytes, size;
  unsigned char *ptr, *ptr1;
 
  int fd;
 
	assert(filename);
	if (!filename) return ERROR;

  fd = open(filename, O_RDWR);
  if (fd == ERROR)
  {
    fprintf(stderr, "Error opening %s: %s\n", filename, strerror(errno));
    return ERROR;
  }
 
  ptr = s_header;
  while (1)
  {
    size = 24;
    ptr1 = ptr;
 
    while (size > 0)
    {
      numbytes = read(fd, ptr1, size);
      if (numbytes <= 0)
      {
        fprintf(stderr, "Error reading header from ASF header: %s\n", strerror(errno));
        return -1;
      }
 
      size -= numbytes;
      ptr1 += numbytes;
    }

		s_asfh = (struct asf_header *)find_guid(guid_header, ptr, 16);
    if (s_asfh != NULL) break;
	
    // see if it's an asf 2.0 header
		s_asfh = (struct asf_header *)find_guid(guid_header2, ptr, 16);
    if (s_asfh != NULL)
		{
			assert(0); // TODO: this is an asf v2.0 header
			break;
		}
		
		ptr += 24;
	}

	s_header_size = s_asfh->size_lo;
	size = s_header_size - 24;
 	
	memmove(s_header, s_asfh, 24); // copy header to the beginning of the buffer
	ptr = s_header + 24;
	
	while (size > 0)
	{
		numbytes = read(fd, ptr, size);
		if (numbytes <= 0) 
		{
			fprintf(stderr, "Error reading header from ASF file: %s\n", strerror(errno));
			return ERROR;
		}

		size -= numbytes;
		ptr += numbytes;
	}
	
	size = s_header_size - 24;

	s_asffh = (struct asf_file_header *)find_guid(guid_file_header, s_header + 24, size);
	s_file_header_size = s_asffh->size_lo;

	return fd;
}

/*
 * asf_read_asx_file
 */
int asf_read_asx_file(char *filename, char *data, int size)
{
	int numbytes;
	unsigned char *ptr;
	
	struct stat stat;
	
	int fd;
 
  assert(filename);
	assert(data);
	assert(size > 0);
	if (!filename) return ERROR;
	if (!data) return ERROR;
	if (size <= 0) return ERROR;

	fd	= open(filename, O_RDONLY); // in case they later write to the file
	if (fd == ERROR)
	{
		fprintf(stderr, "Error opening %s: %s\n", filename, strerror(errno));
		return ERROR;
	}

	if (fstat(fd, &stat) < 0)
	{
		fprintf(stderr, "Error with fstat on %s: %s\n", filename, strerror(errno));
		return ERROR;
	}
	
	if (stat.st_size > size - 1)
	{
		printf("Error in asf_read_asx_file: buffer must be at least %lu bytes long\n", stat.st_size+1);
		return ERROR;
	}
	
	size = stat.st_size;
	ptr = data;
	
	while (size > 0)
	{
		numbytes = read(fd, ptr, size);
		
		if (numbytes <= 0)
		{
			if (numbytes == ERROR && errno == EINTR) continue;

			fprintf(stderr, "Error reading from %s: %s\n", filename, strerror(errno));
			return ERROR;
		}
		
		size -= numbytes;
		ptr += numbytes;
	}
	
	return fd;
}

/*
 * asf_write_header
 */
int asf_write_header(int fd, struct asf_header *asfh, int size)
{
	int numbytes;
	unsigned char *ptr = (unsigned char *)asfh;
	
  assert(fd >= 0);
	assert(size > 0);
	assert(asfh);
	if (fd < 0) return ERROR;
	if (size <= 0) return ERROR;
	if (!asfh) return ERROR;
	
	memcpy(asfh->id, guid_header, 16);
	asfh->size_hi = 0;
	asfh->size_lo = (unsigned long)size;
	
	if (asf_reset_file(fd) < 0) return ERROR; // write the header at the beginning of the file

	while (size > 0)
	{
		numbytes = write(fd, ptr, size);
		if (numbytes <= 0)
		{
			fprintf(stderr, "Error writing header to ASF file: %s\n", strerror(errno));
			return ERROR;
		}

	  size -= numbytes;
		ptr += numbytes;
	}
	
	return 0;
}

/*
 * asf_write_data
 */
int asf_write_data(int fd, unsigned char *data, int size, unsigned long seqnum, unsigned long chunklength)
{
	int numbytes, padding;
	unsigned char *ptr = data;
	
	assert(fd >= 0);
	assert(size > 0);
	assert(chunklength > 0);
	assert(s_header_size > 0);
	assert (data);
	if (fd < 0) return ERROR;
	if (size <= 0) return ERROR;
	if (chunklength <= 0) return ERROR;
	if (s_header_size <= 0) return ERROR;
	if (!data) return ERROR;

	if (lseek(fd, s_header_size + (seqnum * chunklength), SEEK_SET) < 0)
	{
		fprintf(stderr, "Error seeking position 0x%08lx in ASF file: %s\n", s_header_size + (seqnum * chunklength), strerror(errno));
		return ERROR;
	}
	
	padding = size - chunklength;

	while (size > 0)
	{
		numbytes = write(fd, ptr, size);
		if (numbytes <= 0)
		{
			fprintf(stderr, "Error writing data chunk to ASF file (wrote %d of %u bytes): %s\n", numbytes, size, strerror(errno));
			return ERROR;
		}

	  size -= numbytes;
		ptr += numbytes;
	}
	
	while (padding--)
	{
		numbytes = write(fd, 0, 1);
	 	if (numbytes <= 0)
	 	{
	 		fprintf(stderr, "Error writing padding to ASF file (%u bytes left): %s\n", padding, strerror(errno));
	  	return ERROR;
	 	}
	}
	
	return 0;
}

////////////////////////////////////////////////////////////////
// ASF header locaton functions

/*
 * asf_locate_header
 */
struct asf_header *asf_locate_header(int fd)
{
	assert (fd >= 0);
	if (fd < 0) return NULL;

	return s_asfh;
}

/*
 * asf_locate_file_header
 */
struct asf_file_header *asf_locate_file_header(int fd)
{
	assert (fd >= 0);
	if (fd < 0) return NULL;
	
	return s_asffh;
}

/*
 * asf_locate_stream_header
 */
struct asf_stream_header *asf_locate_stream_header(int fd)
{
	static struct asf_stream_header *s_asfsh = NULL;
	
	unsigned char *ptr;
	unsigned long offset;
 
	assert (fd >= 0);
	if (fd < 0) return NULL;
	
	if (s_asfsh) ptr = (unsigned char *)s_asfsh + sizeof(struct asf_stream_header) - sizeof(s_asfsh->data) + sizeof(s_asfsh->data.audio);
	else ptr = (unsigned char *)s_asffh + s_file_header_size;
	
  offset = (unsigned long)ptr - (unsigned long)s_asfh;
  s_asfsh = (struct asf_stream_header *)find_guid(guid_stream_header, ptr, s_header_size - offset);
	
	return s_asfsh;
}

/*
 * asf_locate_data_header
 */
struct asf_data_header *asf_locate_data_header(int fd)
{
	static struct asf_data_header *s_asfdh = NULL;
 
	int numbytes, total_bytes_read = 0;
	
	unsigned char *ptr, *ptr1;
	unsigned long size = s_header_size;
 
 	if (s_asfdh) return s_asfdh;
 
	s_asfdh = (struct asf_data_header *)s_data;
	ptr = (unsigned char *)s_asfdh;
	
	while (1)
	{
		ptr1 = (unsigned char *)ptr;
		size = sizeof(struct asf_data_header);
	 
		while (size > 0)
		{
			numbytes = read(fd, ptr1, size);
			if (numbytes <= 0)
			{
				printf("Error reading data header (read %d of %lu bytes): %s\n", numbytes, size, strerror(errno));
				return NULL;
			}
	 
			size -= numbytes;
			ptr1 += numbytes;
		}
		
		total_bytes_read += sizeof(struct asf_data_header);
	
		s_asfdh = (struct asf_data_header *)find_guid(guid_data_chunk, ptr, total_bytes_read);
		if (s_asfdh) break;
	}
	
	return s_asfdh;
}

////////////////////////////////////////////////////////////////
// ASF parsing/print functions

/*
 * asf_print_packet_header
 */
void asf_print_packet_header(struct asf_packet_header *asfph)
{
  if (!asfph) return;
  printf("PACKET HEADER: %u bytes\n\n", asfph->length);

  if (asfph->length != asfph->length_2)
  {
    fprintf(stderr, "Error: corrupt packet header (length = %u, length_2 = %u)\n\n", asfph->length, asfph->length_2);
    return;
  }

  printf("packet type = $%c\n", (char)((asfph->type >> 8) & 0xff)); 
  printf("packet seqnum = 0x%03lx\n", asfph->seqnum);
	printf("packet flags = 0x%04x", asfph->flags);

	putchar('\n');
}

/*
 * asf_print_header
 */
void asf_print_header(struct asf_header *asfh)
{
  unsigned char *pos, *ptr;
  unsigned long size;

	assert(asfh);
	if (!asfh) return;
	
  size = asfh->size_lo - 24;

	printf("HEADER: %lu bytes\n\n", size + 24);
  if (!size) return;

	printf("unknown 1 = 0x%08lx (%lu)\n", asfh->unknown_1, asfh->unknown_1);
	printf("unknown 2 = 0x%02x 0x%02x\n", asfh->unknown_2[0], asfh->unknown_2[1]);
  putchar('\n');
	
  ptr = (unsigned char *)asfh+24;

	pos = find_guid(guid_bitrates_1, ptr, size);
	if (pos)
	{
		int i;
		unsigned char *ptr1;
		unsigned long bitrate;
		unsigned short num_bitrates, id;

		pos += 24, ptr1 = pos;

		num_bitrates = *((unsigned short *)pos), pos += 2;
		printf("Found 1 audio and %u video streams:\n", num_bitrates - 1);
		
		for (i = 0; i < num_bitrates; i++)
		{
			id = *((unsigned short *)pos), pos += 2;
			bitrate = *((unsigned long *)pos), pos += 4;

			if (id == 0x01)	printf("Stream ID 0x%02x (audio) = %lu\n", id, bitrate);
			else	printf("Stream ID 0x%02x (video) = %lu\n", id, bitrate);
		}

		putchar('\n');
	}
	
#if 0 // guid_bitrates_1 gives more useful information
	pos = find_guid(guid_bitrates_2, ptr, size);
	if (pos)
	{
		int i;
		unsigned short id, num_bitrates;
		
		pos += 16;
		num_bitrates = *((unsigned short *)pos);
		printf("This file has %u bitrates.  Stream IDs: ", num_bitrates);
		
		for (i = 0; i < num_bitrates; i++)
		{
			pos += 2;
			id = *((unsigned short *)pos);
			printf("0x%02x ", id);
		}

		printf("\n\n");
	}
#endif
	
#ifdef SHOW_CONTENT_TYPE
	pos = find_guid(guid_content_type, ptr, size);
	if (pos)
	{
		register int i;
		struct asf_content_type *asfct = (struct asf_content_type *)pos;

		printf("CONTENT TYPE (%lu bytes)\n\n", asfct->size_lo);
		
		printf("unknown id = "), print_hex(asfct->unknown_id, 16, -1);
		printf("unknown = 0x%08lx\n", asfct->unknown);

		pos += sizeof(struct asf_content_type);
				
		printf("\nContent type:\n");
		for (i = 0; i < asfct->size_lo - sizeof(struct asf_content_type); i += 2) 
		{
			if (!pos[i]) putchar('\n');
			if (isalnum(pos[i]) || isspace(pos[i])) putchar(pos[i]);
		}

		printf("\n\ncontent_type[] (%lu bytes) =\n", asfct->size_lo - 24);
		print_hex((unsigned char *)asfct + 24, asfct->size_lo - 24, 0);
	 	putchar('\n');
	}
#endif

	pos = find_guid(guid_url, pos, size);
	if (pos)
	{	
		register int i;
		pos += 36; // skip over guid and the garbage

		printf("Component download URL: ");	
		for (i = 0; pos[i]; i += 2)
		{
			if (isalnum(pos[i]) || ispunct(pos[i])) putchar(pos[i]);
		}

		printf("\n\ncomponent_url[] =\n");
		print_hex(pos - 20, 20 + i, 0);
		putchar('\n');
	}

#if 0 // this doesn't seem to have anything useful in it	
	pos = find_guid(guid_unknown_1, ptr, size);
	if (pos)
	{
		unsigned long len;
	 
		pos += 16;
		len = *((unsigned long *)pos), pos += 8;

		printf("unknown_1[] (%lu bytes) =\n", len);
		print_hex(pos, len - 24, 0);
		putchar('\n');
	}
#endif
	
	printf("\nheader[] (%lu bytes) =\n", size);
  print_hex(ptr, size, 0);
  putchar('\n');
}

/*
 * asf_print_file_header()
 */
void asf_print_file_header(struct asf_file_header *asffh)
{
  unsigned long size;
	unsigned char *ptr;
	double seconds;
	
	assert(asffh);
  if (!asffh) return;
  
  size = asffh->size_lo - 24;
	
	printf("FILE HEADER: %lu bytes\n", size + 24);
  if (!size)
	{
		putchar('\n');
		return;
	}
	
	printf("FILE ID: "), print_hex((unsigned char *)asffh->file_id, 16, -1);
	printf("FILE SIZE: %lu bytes\n\n", asffh->file_size_lo);

	if (size + 24 < sizeof(struct asf_file_header))
	{
		fprintf(stderr, "Error: file header size (%lu) < asf_file_header (%u)\n", size + 24, sizeof(struct asf_file_header));
		return;
	}
	
	if (asffh->chunklength != asffh->chunklength_2)
	{
		printf("Error: corrupt file header (chunklength (0x%08lx) != chunklength_2 (0x%08lx))\n\n", asffh->chunklength, asffh->chunklength_2);
		return;
	}
	
	printf("chunk length = %lu bytes\n\n", asffh->chunklength);

	seconds = ((asffh->file_time_hi * 0x100000000) + asffh->file_time_lo) / 10000000;
	printf("file time = 0x%08lx 0x%08lx (%lu)\n", asffh->file_time_hi, asffh->file_time_lo, asffh->file_time_lo);
	printf("file created %.0f seconds after 01/01/1601 00:00:00 (UTC)\n\n", seconds); // TODO; give this in format of date and hour:min:sec created
	
	printf("unknown 1 = 0x%08lx 0x%08lx (%lu)\n", asffh->unknown_1_hi, asffh->unknown_1_lo, asffh->unknown_1_lo);
	printf("unknown 2 = 0x%08lx 0x%08lx (%lu)\n", asffh->unknown_2_hi, asffh->unknown_2_lo, asffh->unknown_2_lo);
	printf("unknown 3 = 0x%08lx 0x%08lx (%lu)\n", asffh->unknown_3_hi, asffh->unknown_3_lo, asffh->unknown_3_lo);
	printf("unknown 4 = 0x%08lx (%lu)\n", asffh->unknown_4, asffh->unknown_4);
	printf("unknown 5 = 0x%08lx (%lu)\n", asffh->unknown_5, asffh->unknown_5);
	printf("unknown 6 = 0x%08lx (%lu)\n", asffh->unknown_6, asffh->unknown_6);
	printf("unknown 7 = 0x%08lx (%lu)\n", asffh->unknown_7, asffh->unknown_7);
	
	ptr = (unsigned char *)asffh + 24;
	
	printf("\nfile_header[] (%lu bytes) =\n", size);
  print_hex(ptr, size, 0);
	putchar('\n');

	if (size - sizeof(struct asf_file_header) + 24 > 0)
	{
		printf("remainder of file header (%lu bytes) =\n", size - sizeof(struct asf_file_header) + 24);
		print_hex(ptr + sizeof(struct asf_file_header), size - sizeof(struct asf_file_header) + 24, 0);
		putchar('\n');
	}
}

/*
 * asf_print_stream_header
 */
void asf_print_stream_header(struct asf_stream_header *asfsh)
{
  unsigned char *pos, *ptr;
  unsigned long size;

	assert(asfsh);
  if (!asfsh) return;

	size = asfsh->size_lo - 24;
  
	printf("STREAM HEADER: %lu bytes\n", size + 24);
  if (!size)
	{
		putchar('\n');
		return;
	}

	printf("STREAM ID: %u\n\n", asfsh->stream_id);
	
	printf("header length = %lu (0x%08lx)\n", asfsh->datalen_1, asfsh->datalen_1);
	printf("data length = %lu (0x%08lx)\n", asfsh->datalen_2, asfsh->datalen_2);
	printf("unknown_1 = 0x%08lx 0x%08lx (%lu)\n", asfsh->unknown_1_hi, asfsh->unknown_1_lo, asfsh->unknown_1_lo);
	printf("unknown_2 = 0x%08lx (%lu)\n", asfsh->unknown_2, asfsh->unknown_2);
	putchar('\n');
	
	ptr = (unsigned char *)asfsh + 24;
	
	if ((pos = find_guid(guid_audio_stream, ptr, size)) != NULL)
	{
		printf("Audio properties:\n");
		printf("format type = 0x%02x\n", asfsh->data.audio.format_type);
		printf("size = %u (0x%04x)\n", asfsh->data.audio.size, asfsh->data.audio.size);
		printf("number of channels = %u (0x%04x)\n", asfsh->data.audio.num_channels, asfsh->data.audio.num_channels);
		printf("samples per second = %lu (0x%08lx)\n", asfsh->data.audio.samples_per_sec, asfsh->data.audio.samples_per_sec);
		printf("average bytes per second = %lu (0x%08lx)\n", asfsh->data.audio.avg_bytes_per_sec, asfsh->data.audio.avg_bytes_per_sec);
		printf("block size = %u (0x%04x)\n", asfsh->data.audio.block_size, asfsh->data.audio.block_size);
		printf("bits per sample = %u (0x%04x)\n", asfsh->data.audio.bits_per_sample, asfsh->data.audio.bits_per_sample);
	}
	
	else if ((pos = find_guid(guid_video_stream, ptr, size)) != NULL )
	{
#if 1
		printf("Video stream\n");
#else // TODO: these fields don't seem right
		printf("Video properties:\n");
		printf("size = %u (0x%04x)\n", asfsh->data.video.size, asfsh->data.video.size);
		printf("width = %ld (0x%08lx)\n", asfsh->data.video.width, asfsh->data.video.width);
		printf("height = %ld (0x%08lx)\n", asfsh->data.video.height, asfsh->data.video.height);
		printf("unknown = 0x%02x\n", asfsh->data.video.unknown);
		putchar('\n');
		
		printf("Video bitmap properties:\n");
		printf("size = %lu (0x%08lx)\n", asfsh->data.video.bitmap.size, asfsh->data.video.bitmap.size);
		printf("width = %lu (0x%08lx)\n", asfsh->data.video.bitmap.width, asfsh->data.video.bitmap.width);
		printf("height = %lu (0x%08lx)\n", asfsh->data.video.bitmap.height, asfsh->data.video.bitmap.height);
		printf("planes = %u (0x%04x)\n", asfsh->data.video.bitmap.planes, asfsh->data.video.bitmap.planes);
		printf("bit count = %u (0x%04x)\n", asfsh->data.video.bitmap.bit_count, asfsh->data.video.bitmap.bit_count);
		printf("compression = 0x%08lx\n", asfsh->data.video.bitmap.compression);
		printf("image size = %lu (0x%08lx)\n", asfsh->data.video.bitmap.image_size, asfsh->data.video.bitmap.image_size);
		printf("x pels per meter = %ld (0x%08lx)\n", asfsh->data.video.bitmap.x_pels_per_meter, asfsh->data.video.bitmap.x_pels_per_meter);
		printf("y pels per meter = %ld (0x%08lx)\n", asfsh->data.video.bitmap.y_pels_per_meter, asfsh->data.video.bitmap.y_pels_per_meter);
		printf("clr used = 0x%08lx\n", asfsh->data.video.bitmap.clr_used);
		//printf("clr important = 0x%08lx\n", asfsh->data.video.bitmap.clr_important); // TODO: too big
#endif
	}

	else if ((pos = find_guid(guid_unknown_stream, ptr, size)) != NULL)
	{
		printf("Unknown stream (scripting stuff?)\n");
	}

	else assert(0);
	
	putchar('\n');
	
  LOCATE_GUID(guid_audio_stream, ptr, "stream_header", size, "audio stream", 0);
  LOCATE_GUID(guid_video_stream, ptr, "stream_header", size, "video stream", 0);
  LOCATE_GUID(guid_audio_conceal_none, ptr, "stream_header", size, "non-concealed audio", 0);
  LOCATE_GUID(guid_audio_conceal_interleaved, ptr, "stream_header", size, "interleaved audio", 0);
  LOCATE_GUID(guid_video_conceal_none, ptr, "stream_header", size, "non-concealed video", 0);

	LOCATE_GUID(guid_unknown_1, ptr, "header", size, "unknown object 1", 1);
	LOCATE_GUID(guid_unknown_2, ptr, "header", size, "unknown object 2", 1);
  //LOCATE_GUID(guid_unknown_3, ptr, "header", size, "unknown object 3", 1); // subobject of object 2
	
	printf("\nstream_header[] (%lu bytes) =\n", size);
  print_hex(ptr, size, 0);
  putchar('\n');
}

/*
 * asf_print_data_header
 */
void asf_print_data_header(struct asf_data_header *asfdh)
{
	assert(asfdh);
	if (!asfdh) return;

	printf("DATA HEADER (%lu bytes)\n\n", asfdh->size_lo);
	
	printf("packet id: "), print_hex(asfdh->packet_id, 16, -1);
	printf("packet size (number of?) = 0x%08lx 0x%08lx (%lu)\n", asfdh->packet_size_hi, asfdh->packet_size_lo, asfdh->packet_size_lo);

#if 0
	if (!asfdh->chunklength || !asfdh->chunklength_2)
	{
		printf("Erorr: corrupt data header (chunklength = 0x%08lx, chunklength_2 = 0x%08lx)\n\n", asfdh->chunklength, asfdh->chunklength_2);
		return;
	}
	
	if (asfdh->chunklength != asfdh->chunklength_2)
	{
		printf("Error: corrupt data header (chunklength (0x%08lx) != chunklength_2 (0x%08lx))\n\n", asfdh->chunklength, asfdh->chunklength_2);
		return;
	}
	
	printf("total_size 0x%08lx 0x%08lx (%lu)\n", asfdh->total_size_hi, asfdh->total_size_lo);

  printf("unknown_1[] = ");
  print_hex(asfdh->unknown_1, 8, -1);
  putchar('\n');

	printf("numpackets = 0x%08lx 0x%08lx (%lu)\n", asfdh->numpackets_hi, asfdh->numpackets_lo, asfdh->numpackets_lo);
	printf("total_time = 0x%08lx 0x%08lx (%lu)\n", asfdh->total_time_hi, asfdh->total_time_lo, asfdh->total_time_lo);
	printf("play_time = 0x%08lx 0x%08lx (%lu)\n", asfdh->play_time_hi, asfdh->play_time_lo, asfdh->play_time_lo);
	printf("playtime_offset = 0x%08lx (%lu)\n", asfdh->playtime_offset, asfdh->playtime_offset);

  printf("unknown_2[] = ");
	print_hex(asfdh->unknown_2, 4, -1);
  putchar('\n');

	printf("flags = 0x%08lx\n", asfdh->flags);
	printf("chunklength = 0x%08lx, chunklength_2 = 0x%08lx\n", asfdh->chunklength - sizeof(struct asf_data_chunk_header), 
	       asfdh->chunklength_2 - sizeof(struct asf_data_chunk_header));
#endif
	
	printf("\ndata_header[] (%u bytes) =\n", sizeof(struct asf_data_header));
  print_hex((unsigned char *)asfdh, sizeof(struct asf_data_header), 0);
  putchar('\n');
}

////////////////////////////////////////////////////////////////
// Miscellaneous ASF functions

/*
 * asf_get_timecode
 */
unsigned long asf_get_timecode(unsigned char *ptr)
{
  unsigned long timecode;
  unsigned char *timecode_ptr;

	assert(ptr);
  if (!ptr) return 0;

  printf("flags: 0x%02x\n", ptr[3]);

  switch(ptr[3])
  {
    case 0x00:
    case 0x01:
      timecode_ptr = ptr+5;
      break;

    case 0x08:
    case 0x09:
      timecode_ptr = ptr+6;
      break;

    case 0x10:
    case 0x11:
      timecode_ptr = ptr+7;
      break;

    case 0x48:
      timecode_ptr = ptr+8;
      break;

    case 0x50:
      timecode_ptr = ptr+9;
      break;

    default:
			printf("unknown flags = 0x%02x\n", ptr[3]);
			assert(0);
      return 0;
  }

  timecode = *((unsigned long *)timecode_ptr);
#ifdef DEBUG
  printf("timecode = "), print_time(timecode), printf("\n\n");
#endif
	return timecode;
}

/*
 * asf_fix_timecodes
 */
int asf_fix_timecodes(unsigned char *ptr, unsigned long deltatime, unsigned long chunknum, unsigned long chunklength)
{
  unsigned char *ptr1, b1;
  unsigned long f1, f2, l1, l2, l3, l4, l5, l6, l7, l8, n1, n;
  unsigned long v1, v2, v3, v4, v5 = 0, v6, v7, v8, v9, v10;
  
	if (!(ptr[0] == 0x82 && ptr[1] == 0x00 && ptr[2] == 0x00))
  {
#ifdef DEBUG
    printf("Error: illegal header bytes in chunk %ld!\n", chunknum);
#endif
    return 1;
  }
	
  ptr1 = &ptr[3];
  b1 = *ptr1++;
  f1 = (b1 & 0x80) >> 7;
  
	if (f1 == 0)
  {
    l1 = (b1 & 0x60) >> 5;
    l2 = (b1 & 0x18) >> 3;
    l3 = (b1 & 0x06) >> 1;
    f2 = (b1 & 0x01);
  }
  
	else
  {
#ifdef DEBUG
    printf("Error: chunks carrying error correction information are not supported yet!\n");
#endif
		return 1;
  }
  
	b1 = *ptr1++;
  l4 = (b1 & 0xc0) >> 6;
  l5 = (b1 & 0x30) >> 4;
  l6 = (b1 & 0x0c) >> 2;
  l7 = (b1 & 0x03);
  v1 = get_data(&ptr1, l1);
  v2 = get_data(&ptr1, l3);
  v3 = get_data(&ptr1, l2);
	
  v4 = get_data(&ptr1, 3);
  v4 = (deltatime < v4) ? v4 -= deltatime : 0;
  put_long(ptr1 - 4, v4);

	if (l1 == 0) v1 = chunklength;
  if (l3 == 0) v2 = chunknum;
  if (v3 > 0) memset(ptr + v1 - v3, 0, v3);

	ptr1 += 2;
  b1 = *ptr1++;
  //l8 = (b1 & 0x80) >> 7;
  l8 = 0;
  n1 = (b1 & 0x1f);
  
	if (f2 == 0)
  {
    n1 = 1;
    l8 = 0;
  }
	
  for (n = 0; (n < n1); n++)
  {
    if (f2 != 0) v5 = get_data(&ptr1, l4);
    v6 = get_data(&ptr1, l5);
    v7 = get_data(&ptr1, l6);
    v8 = get_data(&ptr1, l7);
    if (v8 == 8)
    {
      get_long(ptr1 + 4, &v10);
      v10 = (deltatime < v10) ? v10 -= deltatime : 0;
      put_long(ptr1 + 4, v10);
    }
    if (v8 > 0) ptr1 += v8;

		if (f2 != 0) v9 = get_data(&ptr1, l8);
    else v9 = (v1 - v3) - (ptr1 - ptr);

		if (v9 > 0) ptr1 += v9;
  
		if ((unsigned long)(ptr1 - ptr) > chunklength)
    {
#ifdef DEBUG
      printf("Warning: ptr points %ld bytes after packet!\n", (unsigned long)(ptr1 - ptr) - chunklength);
#endif
      return 1;
    }
    if ((unsigned long)(ptr1 - ptr) > chunklength - v3)
    {
#ifdef DEBUG
      printf("Warning: ptr pointer %ld bytes after packet-padlen!\n", (unsigned long)(ptr1 - ptr) - (chunklength - v3));
#endif
      return 1;
    }
  }
	
#ifdef DEBUG
  if ((ptr1 - ptr) < (unsigned long)chunklength - (unsigned long)v3)
		printf("Note: %ld bytes not covered by payload\n", chunklength - v3 - (ptr1 - ptr));
#endif
	
	return 0;
}
