/* 
 * Usage: asfreceiver <output ASF file> <media_server:port> <URL> [proxy_server:port] [username] [password]
 * Example: asfreceiver output.asf media_server.com:1080 /media_stream.asf
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include "util/util_all.h"
#include "net/net_all.h"

#include "asf.h"
#include "misc.h"
#include "guids.h"

#define MAX_SEQNUMS 16

// Options supported:
//#define DEBUG
//#define QUIET
#define STATISTICS
#define ADJUST_TIMING_OFFSET // causes it to update timecodes as they come

#if defined(DEBUG) && defined(QUIET)
#error "DEBUG and QUIET are mutually exclusive"
#endif

#define PLAYER_VERSION "NSPlayer/4.1.0.3856"

unsigned char new_stream = 0, new_seqnum = 0, no_subchunks = 0;
unsigned char header_found = 0, stream_found = 0;

unsigned long numchunks = 0;

unsigned long seqnum_index = 0;

#ifdef STATISTICS
unsigned short min_packet_size, max_packet_size = 0;
#endif

unsigned int stream_off_hi = 0, stream_off_lo = 0;
unsigned int max_duration = 0, max_timecode = 0;
unsigned long start_seqnum = 0, start_time = 0;

char file[256];
char server[128], url[256], proxy_server[128], username[64], password[64];

int resume_file = 0;
int live_stream = 0;

struct seqnum
{
  unsigned short value;
  unsigned short beginning_seqnum;
  unsigned short count;
  unsigned short subchunk_length, data_length;
  unsigned short type;
  unsigned char drop, asked;
} seqnums[MAX_SEQNUMS];

void set_auth_params()
{
  char *ptr;

  printf("HTTP Authorization\n");
	
	printf("Username (blank if none): ");
	fgets(username, sizeof(username)-1, stdin);
 
	for (ptr = username; *ptr; ptr++)
	{
		if (*ptr == '\r' || *ptr == '\n')
		{
			*ptr = '\0';
			break;
		}
	}
	
	if (isalnum(username[0]))
	{
		printf("Password (blank if none): ");
	  fgets(password, sizeof(password)-1, stdin);
	
		for (ptr = password; *ptr; ptr++)
		{
			if (*ptr == '\r' || *ptr == '\n')
			{
				*ptr = '\0';
				break;
			}
		}

  	putchar('\n');
	}
}

int main(int argc, char **argv)
{
	struct guid guid;
  struct asf_packet_header *asfph = NULL;
  struct asf_header *asfh = NULL;
  struct asf_file_header *asffh = NULL;
	struct asf_stream_header *asfsh = NULL;
	struct asf_data_header *asfdh = NULL;
	
  int fd = -1, sock;

  int numbytes, offset;
  unsigned long	header_offset = 0, body_length = 0, header_length = 0;
	
	int found_stream = 0;
	
	unsigned long i;
	unsigned long timecode = 0;
	unsigned long size = 0, header_size = 0;

  unsigned char *ptr, *pos;
  static unsigned char buf[BUFSIZE], request[1024], content_type[32];

	struct ipport *ipp;

	unsigned long file_size_hi, file_size_lo;
	unsigned long segsize_hi, segsize_lo;

	double total_time;
	unsigned long total_time_hi, total_time_lo;

	double playtime;
	unsigned long playtime_hi, playtime_lo;

	unsigned long numchunks_hi, numchunks_lo;

	// Deal with parameters

  if (argc < 4 || argc > 5)
  {
     fprintf(stderr, "Usage: %s <output ASF file> <media_server:port> <URL> [proxy_server:port]\n", argv[0]);
		 fprintf(stderr, "Example: %s output.asf media_server.com:1080 /media_stream.asf\n", argv[0]);

		 return 0;
  }

	set_auth_params(); // setup HTTP auth params

	strncpy(file, argv[1], sizeof(file)-1);
	strncpy(server, argv[2], sizeof(server)-1);
	strncpy(url, argv[3], sizeof(url)-1);

	if (argc == 5) strncpy(proxy_server, argv[4], sizeof(proxy_server)-1);

	// Avoid crashes due to unnecessarily long arguments

	if (strlen(url) >= 128)
	{
		fprintf(stderr, "URL must be < 128 characters\n");
		return ERROR;
	}

	if (strlen(proxy_server) >= 128)
	{
		fprintf(stderr, "Proxy host must be < 128 characters\n");
		return ERROR;
	}
	
	if (strlen(username) >= 64)
	{
		fprintf(stderr, "Username must be < 64 characters\n");
		return ERROR;
	}
	
	if (strlen(password) >= 64)
	{
		fprintf(stderr, "Password must be < 64 characters\n");
		return ERROR;
	}
  
	
  // Resume or overwrite file

	if (access(file, F_OK) == 0)
	{
		resume_file = 1;

		fd = open(file, O_RDONLY);
		if (fd < 0)
		{
			fprintf(stderr, "\nError opening %s: %s\n", file, strerror(errno));
			return -1;
		}
		
		stream_off_lo = lseek(fd, 0, SEEK_END);
		close(fd);

		if (stream_off_lo <= sizeof(struct asf_header) + sizeof(struct asf_file_header) + sizeof(struct asf_stream_header))
		{
			fprintf(stderr, "\nUnable to resume %s (incomplete file--remove it)\n", file);
			return -1;
		}
		
		fd = asf_append_file(file);
	}
	
	else fd = asf_create_file(file); 
	
  if (fd == ERROR) return ERROR;

  // Initialize net stuff	
	net_init(0, 0);
	
	ipp = net_hostname_to_ipport(server);
	if (!ipp)
	{
		fprintf(stderr, "\nError resolving %s (in hostname:port format?)\n", server);
		return ERROR;
	}
	
	///////////////////////////////////////////////////////////////////
	// PHASE 1
	///////////////////////////////////////////////////////////////////
	
	printf ("\nConnecting to %s...\n", net_ip_to_string(ipp->ipa));

	sock = net_initiate_tcp_connect(ipp->ipa, ipp->nsport);
	if (sock < 0)
	{
		fprintf(stderr, "Error connecting to %s: %s\n", net_ip_to_string(ipp->ipa), strerror(errno));
		return ERROR;
	}

	printf("Successfully connected to %s (phase 1)\n\n", net_ip_to_string(ipp->ipa));
	
	guid = guid_make_random_guid();

	// Setup URL

	ptr = request;
	
	if (!isalnum(proxy_server[0])) ptr += sprintf(ptr, "GET %s HTTP/1.0\r\n", url);
	else ptr += sprintf(ptr, "GET http://%s/%s HTTP/1.0\r\n", server, url);
	
  if (isalnum(username[0])) ptr += sprintf(ptr, "Authorization: Basic %s\r\n", base64enc("%s:%s", username, password));
  ptr += sprintf(ptr, "Accept: */*\r\n");
  ptr += sprintf(ptr, "User-Agent: %s\r\n", PLAYER_VERSION);
  if (isalnum(proxy_server[0])) ptr += sprintf(ptr, "Host: %s\r\n", proxy_server);
  
	sprintf(ptr,
	        "Pragma: no-cache,rate=1.000000,stream-time=0,stream-offset=0:0,request-context=1,max-duration=0\r\n"
	        "Pragma: xClientGUID={%s}\r\n"
	        "Connection: close\r\n"
				  "\r\n",
	        guid_to_string_win_format(guid));

#ifdef DEBUG
	printf("Sending:\n%s", ptr);
#endif
	
	numbytes = net_send(sock, request, strlen(request));
	if (numbytes < 0)
	{
		fprintf(stderr, "Failed to send request (sent %d of %d bytes): %s\n", numbytes, strlen(request), strerror(errno));
		return ERROR;
	}
	
	memset(buf, 0, sizeof(buf));
	asfph = (struct asf_packet_header *) buf;  

	sprintf(content_type, "application/octet-stream"); // default
	
#ifdef DEBUG
	printf("Receiving the following data:\n");
#endif
	
	while (1)
	{
		// TODO: read in bigger steps to catch HTTP error and content type

#if 0
		
		ptr = strstr(buf, "Content-Type: ");
		if (!ptr) ptr = strstr(buf, "Content-type: ");
		if (!ptr) ptr = strstr(buf, "content-type: ");

		if (ptr) // if we don't have it now, give up
		{
			ptr += 14; // skip over "Content-Type: "
			snprintf(content_type, sizeof(content_type), "%s", ptr);
			printf("Content type is \"%s\"\n", content_type);
		}
		
		// check if it's redirect content
		if (strcasecmp(content_type, "audio/x-ms-wax") == 0 || strcasecmp(content_type, "audio/x-ms-wma") == 0 || strcasecmp(content_type, "audio/x-ms-asf") == 0 ||
				strcasecmp(content_type, "audio/x-ms-afs") == 0 || strcasecmp(content_type, "video/x-ms-wvx") == 0 || strcasecmp(content_type, "video/x-ms-wmv") == 0 ||
				strcasecmp(content_type, "video/x-ms-wma") == 0)
		{
			numbytes = net_recv(csock, buf);
			if (numbytes <= 0)
			{
				fprintf(stderr, "Error receiving data (read %d bytes): %s\n", numbytes, strerror(errno));
				goto closeup;
			}

#ifdef DEBUG
			printf("Received:\n%s", buf);
#endif

			// TODO: handle redirected content here
			
			goto closeup; // TODO: remove
		}
#endif

		// check for errors 
		ptr = strstr(buf, "HTTP/");
		if (ptr)
		{
			ptr += 9; // skip over "HTTP/X.X "
		
			if (isdigit(*ptr)) // an error occurred
			{
				int error_code = atoi(ptr);
			
				printf("Error code %u occurred\n", error_code);
				goto closeup;
			}
		}
		
		numbytes = net_recv_exact(sock, buf, 1);
		if (numbytes != 1)
		{
			 if (!numbytes) goto closeup;

			 fprintf(stderr, "Error receiving data (%d of 1 byte): %s\n", numbytes, strerror(errno));
			 return ERROR;
		}
		
		else if (buf[0] != '$')
		{
#ifdef DEBUG
			putchar(buf[0]);
#endif
			continue;
		}
		
find_packet1:
		numbytes = net_recv_exact(sock, buf+1, 1); // 'D', 'E', or 'H'
		if (numbytes != 1)
		{
			 fprintf(stderr, "Error receiving data (%d of 1 byte): %s\n", numbytes, strerror(errno));
			 return ERROR;
		}

		if (buf[1] == 'H' && !header_found)
		{
			 if (asfph->type != 0x4824 || asfph->length != asfph->length_2) continue;
			 header_found = 1;
			 break;
		}
		else if (buf[1] == 'D' || buf[1] == 'E') break;

		if (buf[1] == '$') goto find_packet1;
		else continue;

		break;
	}

	numbytes = net_recv_exact(sock, buf+2, 2); // read packet length
	if (numbytes != 2)
	{
		fprintf(stderr, "Error receiving data (%d of 2 bytes): %s\n", numbytes, strerror(errno));
		return ERROR;
	}

	if (asfph->length > BUFSIZE)
	{
		fprintf(stderr, "packet length (%u) > bufsize (%d)\n", asfph->length, sizeof(buf));
		return ERROR;
	}

	body_length = asfph->length - 8;

	numbytes = net_recv_exact(sock, buf + header_offset + 4, asfph->length); // get packet
	if (numbytes != asfph->length)
	{
		fprintf(stderr, "Error receiving data (%d of %u bytes): %s\n", numbytes, asfph->length, strerror(errno));
		return ERROR;
	} 
	
	if (asfph->length != asfph->length_2)
	{
		fprintf(stderr, "Error: corrupt packet header (length = %u, length_2 = %u)\n\n", asfph->length, asfph->length_2);
		return ERROR;
	} 

	body_length += header_offset;

#ifdef DEBUG	
	asf_print_packet_header(asfph);
#endif
	
	if (asfph->type == 0x4824 && !resume_file) // $H
  {
	 	pos = find_guid(guid_header, buf+12, asfph->length - 20);
   	if (!pos)
   	{
	     fprintf(stderr, "Error: corrupt $H packet (no header found)\n\n");
 	     return ERROR;
   	}

	 	if (asfph->flags & 0x0400) // header is split into parts
		{

			header_offset = 0;
			get_long(buf + 28, &header_length);
			header_length += 50;
			
#ifdef DEBUG
			printf("header_length = %lu (0x%08lx)\n", header_length, header_length);
#endif

			// TODO: if this occurs, is it sitll safe to parse header, file header, and stream headers?
			assert(0);
		}
			
		if (!(asfph->flags & 0x0800)) // header hasn't been completely received
		{
			header_offset = body_length;
			body_length = header_length;

#ifdef DEBUG
			printf("body_length = %lu (0x%08lx)\n", body_length, body_length);
#endif
			
			// TODO: if this occurs, is it sitll safe to parse header, file header, and stream headers?
			assert(0); // TODO
		}
   	
		// Header

		asfh = (struct asf_header *)pos;

		// sanity check header size
   	header_size = asfh->size_lo;
   	if (header_size > asfph->length - 8)
   	{
     	fprintf(stderr, "Error: corrupt $H packet (header length (%lu) > packet length (%u))\n", header_size, asfph->length - 8); 
			return ERROR;
   	}

#ifdef DEBUG
   	asf_print_header(asfh);
#endif

		// File header

		pos = find_guid(guid_file_header, buf+12, asfph->length - 20);
   	if (!pos)
   	{
     	fprintf(stderr, "Error: corrupt $H packet (no file header found)\n\n");
     	return ERROR;
   	}
			
		asffh = (struct asf_file_header *)pos;
		size = asffh->size_lo;
		
		if (size > header_size)
		{
			fprintf(stderr, "Error: corrupt $H packet (file header length (%lu) > packet length (%u))\n", size, asfph->length - 8);
			return ERROR;
		}
		
#ifdef DEBUG
   	asf_print_file_header(asffh);
#endif

		// Stream header

		pos = (unsigned char *)asffh + size;
		
		while (1)
		{
			pos = find_guid(guid_stream_header, pos, asfph->length - 20);
   		if (!pos) break;
   		
			found_stream = 1;
			
			asfsh = (struct asf_stream_header *)pos;
			size = asfsh->size_lo;

			if (size > header_size)
			{
				fprintf(stderr, "Error: corrupt $H packet (file header length (%lu) > packet length (%u))\n", size, asfph->length - 8);
				return ERROR;
			}
			
			// TODO: do stream header stuff
			asf_print_stream_header(asfsh);
			
			pos += size; 
		}
		
		if (!found_stream)
		{
    	fprintf(stderr, "Error: corrupt $H packet (no stream headers found)\n\n");
     	return ERROR;
   	}
			
#if 0		
		// TODO: locate data header
		
		if (asfsh->chunklength != asfsh->chunklength_2)
		{
	  	printf("Error: corrupt $H packet (chunklength = %lu, chunklength_2 = %lu)\n\n",
	         	 asfsh->chunklength, asfsh->chunklength_2); 
			return ERROR;
		}
			
		if (!(asfsh->playtime_hi + asfsh->playtime_lo))
		{
			live_stream = 1;
		 	max_duration = 0;
			if (!start_seqnum) start_seqnum = asfph->seqnum;
		}
	  else max_duration = (int)((double)429496.7296 * asfsh->playtime_hi) + (asfsh->playtime_lo / 10000) - asfsh->playtime_offset;
#endif

		// Write header to disk
		if (asf_write_header(fd, asfh, size) < 0) return ERROR;
		
#ifdef DEBUG
    asf_print_stream_header(asfsh);
    putchar('\n');
#endif
  }
	
	net_close_and_remove(sock);

	///////////////////////////////////////////////////////////////////
	// PHASE 2
	///////////////////////////////////////////////////////////////////

	printf("\nReconnecting to %s...\n", net_ip_to_string(ipp->ipa));

	sock = net_initiate_tcp_connect(ipp->ipa, ipp->nsport);
	if (sock < 0)
	{
		fprintf(stderr, "Error reconnecting to %s: %s\n", net_ip_to_string(ipp->ipa), strerror(errno));
		return ERROR;
	}
		
	printf("Successfully reconnected to %s (phase 2)\n\n", net_ip_to_string(ipp->ipa));

	// Setup URL

	ptr = request;
  
	if (!isalnum(proxy_server[0])) ptr += sprintf(ptr, "GET %s HTTP/1.0\r\n", url);
  else ptr += sprintf(ptr, "GET http://%s/%s HTTP/1.0\r\n", server, url);
	
	if (isalnum(username[0])) ptr += sprintf(ptr, "Authorization: Basic %s\r\n", base64enc("%s:%s", username, password));
  ptr += sprintf(ptr, "Accept: */*\r\n");
  ptr += sprintf(ptr, "User-Agent: %s\r\n", PLAYER_VERSION);
  if (isalnum(proxy_server[0])) ptr += sprintf(ptr, "Host: %s\r\n", proxy_server);
  
	sprintf(ptr,
	        "Pragma: no-cache,rate=1.000000,stream-time=0,stream-offset=%u:%u,request-context=2,max-duration=%u\r\n"
				  "Pragma: xPlayStrm=1\r\n"
	        "Pragma: xClientGUID={%s}\r\n"
				  "Pragma: stream-switch-count=1\r\n"
				  "Pragma: stream-switch-entry=ffff:1:0\r\n"
				  "\r\n",
	        stream_off_hi, stream_off_lo, max_duration, guid_to_string_win_format(guid));

#ifdef DEBUG
	printf("Sending:\n%s", ptr);
#endif

	numbytes = net_send(sock, request, strlen(request));
	if (numbytes < 0)
	{
		fprintf(stderr, "Failed to send request (sent %d of %d bytes): %s\n", numbytes, strlen(request), strerror(errno));
		return ERROR;
	}
 	
#ifdef DEBUG
	printf("Receiving the following data:\n");
#endif
	
	while (1)
  {
    memset(buf, 0, sizeof(buf));
    asfph = (struct asf_packet_header *) buf;  

    while (1)
    {
			// TODO: we assume that if it succeeded the first time, it will this time also

			numbytes = net_recv_exact(sock, buf, 1);
      if (numbytes != 1)
      {
         if (!numbytes) goto closeup;

         fprintf(stderr, "Error receiving data (%d of 1 byte): %s\n", numbytes, strerror(errno));
         return ERROR;
      }
      else if (buf[0] != '$') continue;
 
find_packet2:
      numbytes = net_recv_exact(sock, buf+1, 1); // 'D', 'E', or 'H'
      if (numbytes != 1)
      {
         fprintf(stderr, "Error receiving data (%d of 1 byte): %s\n", numbytes, strerror(errno));
         return ERROR;
      }
 
      if (buf[1] == 'H' && !header_found)
      {
         if (asfph->type != 0x4824 || asfph->length != asfph->length_2) continue;
         header_found = 1;
         break;
      }
      else if (buf[1] == 'D' || buf[1] == 'E') break;
 
      if (buf[1] == '$') goto find_packet2;
      else continue;
 
      break;
    }

    numbytes = net_recv_exact(sock, buf+2, 2); // read packet length
    if (numbytes != 2)
    {
      fprintf(stderr, "Error receiving data (%d of 2 bytes): %s\n", numbytes, strerror(errno));
      break;
    }

    if (asfph->length > BUFSIZE)
    {
      fprintf(stderr, "Error: packet length (%u) > bufsize (%d)\n", asfph->length, sizeof(buf));
      continue;
    }

    numbytes = net_recv_exact(sock, buf+4, asfph->length); // get packet
    if (numbytes != asfph->length)
    {
      fprintf(stderr, "Error receiving data (%d of %u bytes): %s\n", numbytes, asfph->length, strerror(errno));
      break;
    }                                                      

  	if (asfph->length != asfph->length_2)
  	{
   		fprintf(stderr, "Error: corrupt packet header (length = %u, length_2 = %u)\n\n", asfph->length, asfph->length_2);
  	  break;
	  } 

#ifdef DEBUG
    asf_print_packet_header(asfph); 
#endif

    /////////////////////////////////////////////////////////////////////////////////

		if (asfph->type == 0x4824)
		{
#ifdef DEBUG
			printf("\n$H received (ignoring)\n");
#endif
			continue; // $H (already read in phase 1)
		}
		
		else if (asfph->type == 0x4524) // $E
    {
#ifdef DEBUG
      printf("\n$E received\n");
#endif
			goto closeup;
    }

    else if (asfph->type == 0x4424) // $D
    {
      unsigned char num_subchunks = 0;
      unsigned char *pos1 = NULL, *endpos = NULL;

      struct asf_data_chunk_header *asfdch = NULL;
      struct asf_data_subchunk_header *asfdsh = NULL;

#ifdef STATISTICS
			if (!min_packet_size || asfph->length < min_packet_size) min_packet_size = asfph->length;
      else if (!max_packet_size || asfph->length > max_packet_size) max_packet_size = asfph->length;
#endif
			
      new_stream = 1;

      asfdch = (struct asf_data_chunk_header *)(asfph + 1);

      if (asfdch->type != 0x82 && asfdch->type != 0x80)
      {
         fprintf(stderr, "Error: unknown $D data block header (type = 0x%02x)\n\n", asfdch->type);
#ifdef DEBUG
				 goto dump_data;
#else
				 continue;
#endif
      }

      pos = (unsigned char *)asfdch;

#ifdef DEBUG
      timecode = asf_get_timecode(pos);
#endif
			if (timecode > max_timecode) max_timecode = timecode;
			if (!start_time) start_time = timecode;
			
      switch (asfdch->flags)
      {
          case 0x00:
          case 0x01:
            pos1 = pos+10;
            break;

          case 0x08:
          case 0x09:
            pos1 = pos+11;
            break;

          case 0x10:
          case 0x11:
            pos1 = pos+12;
            break;

          case 0x48:
            pos1 = pos+13;
            break;

          case 0x50:
            pos1 = pos+14;
            break;

          default:
#ifdef DEBUG
            fprintf(stderr, "Error: unknown or invalid $D packet (data type = 0x%02x)\n", asfdch->flags);
            goto dump_data;
#else
						goto restart_loop;
#endif
			}

      endpos = buf + asfph->length + 3 - sizeof(struct asf_data_subchunk_header);

      no_subchunks = 1;

      for (; pos1 <= endpos; pos1++)
      {
        if (memcmp(pos1+3, "\00\00\00\00\x08", 5)) continue; // yes, a hack

        asfdsh = (struct asf_data_subchunk_header *)pos1;
/*
  			if (asfdsh->data_length > asfdsh->subchunk_length)
  			{
   				fprintf(stderr, "Error: corrupt subchunk (data length (%u) > total subchunk length (%u))\n", asfdsh->data_length, asfdsh->subchunk_length);
   				continue;
			  }
*/
				
#ifdef DEBUG
			  printf("subchunk %u\n", num_subchunks+1);
				asf_print_data_subchunk_header(asfdsh);
#endif

        if (asfdsh->magic_num != 0x08)
        {
#ifdef DEBUG
          fprintf(stderr, "Error: corrupt subchunk (magic number is %02x and not 0x08)\n\n", asfdsh->magic_num);
#endif
          continue;
        }

        num_subchunks++;
        no_subchunks = 0;

        for (i = 0, new_seqnum = 1; i < seqnum_index; i++)
        {
#ifdef DEBUG // match up our sequence numbers
          switch (asfdsh->type)
          {
            case 0x01:
              if (seqnums[i].type != 0x01) continue;

              if (asfdsh->seqnum < seqnums[i].value)
              {
                printf("[type 0x%02x] possible wrap around (current seqnum: 0x%02x, previous seqnum: 0x%02x)\n", asfdsh->type,
                       asfdsh->seqnum, seqnums[i].value);
              }

              else if (asfdsh->seqnum - seqnums[i].value > 1)
              {
                //printf("[type 0x01] comparing 0x%02x to 0x%02x (offset of %u)\n", asfdsh->seqnum, seqnums[i].value,
                //       asfdsh->seqnum - seqnums[i].value);
                continue;
              }

              break;

            case 0x02:
            case 0x82:
              if (seqnums[i].type != 0x02 && seqnums[i].type != 0x82) continue;

              printf("1 bitrate detected\n");

              if (asfdsh->seqnum < seqnums[i].value)
              {
                printf("[type 0x%02x] possible wrap around (current seqnum: 0x%02x, previous seqnum: 0x%02x)\n", asfdsh->type,
                       asfdsh->seqnum, seqnums[i].value);
              }

              break;

            case 0x04:
            case 0x84:
              if (seqnums[i].type != 0x04 && seqnums[i].type != 0x84) continue;

              printf("1 bitrate detected (created as multiple, however)\n");

              if (asfdsh->seqnum < seqnums[i].value)
              {
                printf("[type 0x%02x] possible wrap around (current seqnum: 0x%02x, previous seqnum: 0x%02x)\n", asfdsh->type,
                       asfdsh->seqnum, seqnums[i].value);
              }

              break;

            case 0x05:
            case 0x85:
              if (seqnums[i].type != 0x05 && seqnums[i].type != 0x85) continue;

              printf("2 bitrates detected\n");

              if (asfdsh->seqnum < seqnums[i].value)
              {
                printf("[type 0x%02x] possible wrap around (current seqnum: 0x%02x, previous seqnum: 0x%02x)\n", asfdsh->type,
                       asfdsh->seqnum, seqnums[i].value);
              }

              break;

            case 0x06:
            case 0x86:
              if (seqnums[i].type != 0x06 && seqnums[i].type != 0x86) continue;

              printf("3 bitrates detected\n");

              if (asfdsh->seqnum < seqnums[i].value)
              {
                printf("[type 0x%02x] possible wrap around (current seqnum: 0x%02x, previous seqnum: 0x%02x)\n", asfdsh->type,
                       asfdsh->seqnum, seqnums[i].value);
              }

              break;

            case 0x07:
            case 0x87:
              if (seqnums[i].type != 0x07 && seqnums[i].type != 0x87) continue;

              printf("4 bitrates detected\n");

              if (asfdsh->seqnum < seqnums[i].value)
              {
                printf("[type 0x%02x] possible wrap around (current seqnum: 0x%02x, previous seqnum: 0x%02x)\n", asfdsh->type,
                       asfdsh->seqnum, seqnums[i].value);
              }

              break;

            case 0x08:
            case 0x88:
              if (seqnums[i].type != 0x08 && seqnums[i].type != 0x88) continue;

              printf("5 bitrates detected\n");

              if (asfdsh->seqnum < seqnums[i].value)
              {
                printf("[type 0x%02x] possible wrap around (current seqnum: 0x%02x, previous seqnum: 0x%02x)\n", asfdsh->type,
                       asfdsh->seqnum, seqnums[i].value);
              }

              break;

            default:
              fprintf(stderr, "unknown subchunk type found (type 0x%02x)\n", asfdsh->type);
              continue;
          }
#endif // DEBUG
					
#ifdef STATISTICS
          new_seqnum = 0;

					seqnums[i].value = asfdsh->seqnum;
          seqnums[i].subchunk_length = asfdsh->subchunk_length;
          seqnums[i].data_length = asfdsh->data_length;
          seqnums[i].type = asfdsh->type;
          seqnums[i].count++;
#endif // STATISTICS
					
#ifdef DEBUG
          printf("this subchunk (type 0x%02x) is part of stream %lu (%u packet(s))\n\n", asfdsh->type, i+1, seqnums[i].count);
#endif
          break;
        }

#ifdef STATISTICS
				if (new_seqnum)
        {
          printf("new sequence number detected: seqnum = 0x%02x, type = 0x%02x (stream %lu)\n", asfdsh->seqnum, 
                 asfdsh->type, seqnum_index+1);

          new_seqnum = 0;

          seqnums[seqnum_index].count++;
          seqnums[seqnum_index].value = seqnums[seqnum_index].beginning_seqnum = asfdsh->seqnum;
          seqnums[seqnum_index].subchunk_length = asfdsh->subchunk_length;
          seqnums[seqnum_index].data_length = asfdsh->data_length;
          seqnums[seqnum_index].type = asfdsh->type;
        }
#endif // STATISTICS
				
        pos1 += asfdsh->subchunk_length + sizeof(struct asf_data_subchunk_header) - 2; // skip over the chunk
      }
			
      if (no_subchunks)
      {
#ifdef DEBUG
        printf("no subchunks detected\n");
#endif

#ifdef DEBUG
        printf("\ndata[] =\n");
        print_hex(pos, asfph->length - 8, 0);
        putchar('\n');
#endif
      }

#ifdef DEBUG
      printf("%u total bytes being written for packet\n\n", asfph->length);
#endif

#ifdef ADJUST_TIMING_OFFSET
			if (live_stream)
			{
			  timecode -= start_time;
				asf_fix_timecodes(pos, start_time, asfph->seqnum, asfdh->chunklength); // enables file seeking
			}
#endif

			numbytes = asf_write_data(fd, pos, asfph->length - 8, asfph->seqnum, asfdh->chunklength);
      if (numbytes != asfph->length - 8)
      {
        fprintf(stderr, "Error: failed to write data chunk to the ASF file (wrote %d of %d bytes): %s\n", numbytes, asfph->length - 8, strerror(errno));
        return ERROR;
      }

			numchunks++;
			
#ifdef DEBUG
dump_data:
      printf("\ndata[] =\n");
      print_hex(pos, asfph->length - 8, 0);
      putchar('\n');
#else
restart_loop:
      continue;
#endif
    }
	
	  else
	  {
#ifdef DEBUG
		  fprintf(stderr, "Error: unknown packet data type 0x%02x\n\n", asfph->type);
#endif
		  continue;
    }
	}

	// TODO: write asf_data_header to disk as alternative method

	file_size_hi = 0;
	file_size_lo = lseek(fd, 0, SEEK_END);
	if (file_size_lo < asfph->length - 8)
	{
		fprintf(stderr, "filesize (%lu bytes) is < header size (%u bytes)\n", file_size_lo, asfph->length - 8);
		goto closeup;
	}

#ifdef DEBUG
	printf("ASF file size is %lu bytes\n", file_size_lo); // TODO: remove
#endif
	
	if (lseek(fd, HDR_TOTAL_SIZE_8, SEEK_SET) < 0) goto closeup;
	if (write_quad(fd, file_size_hi, file_size_lo) < 0) goto closeup;
	
	numchunks_hi = 0;
	numchunks_lo = numchunks;
	if (lseek(fd, HDR_NUM_PACKETS_8, SEEK_SET) < 0) goto closeup;
	if (write_quad(fd, numchunks_hi, numchunks_lo) < 0) goto closeup;
 
	total_time = (double)(max_timecode) * 10000;
	total_time_hi = (int)(total_time / 4294967296.0);
	total_time_lo = (int)(total_time - ((double)total_time_hi * 4294967296.0));
	if (lseek(fd, HDR_FINE_TOTALTIME_8, SEEK_SET) < 0) goto closeup;
	if (write_quad(fd, total_time_hi, total_time_lo) < 0) goto closeup;
 
	playtime = (double)(max_timecode - asfdh->playtime_offset) * 10000;
	playtime_hi = (int)(playtime / 4294967296.0);
	playtime_lo = (int)(playtime - ((double)playtime_hi * 4294967296.0));
	if (lseek(fd, HDR_FINE_PLAYTIME_8, SEEK_SET) < 0) goto closeup;
	if (write_quad(fd, playtime_hi, playtime_lo) < 0) goto closeup;
 
	if (lseek(fd, HDR_FLAGS_4, SEEK_SET) < 0) goto closeup;
	if (write_long(fd, 0x00000002) < 0) goto closeup; // enable file seeking

	offset = asfph->length - 8 - DATSEG_HDR_SIZE;

	segsize_hi = 0;
	segsize_lo = DATSEG_HDR_SIZE + asfdh->chunklength;
	if (lseek(fd, offset + DATSEG_SEGSIZE_8, SEEK_SET) < 0) goto closeup;
	if (write_quad(fd, segsize_hi, segsize_lo) < 0) goto closeup;
 
	if (lseek(fd, offset + DATSEG_NUMCHUNKS_4, SEEK_SET) < 0) goto closeup;
	if (write_long(fd, numchunks) < 0) goto closeup;

closeup:

#ifdef STATISTICS
#ifdef DEBUG
  putchar('\n');
#endif
	
  for (i = 0; i < seqnum_index; i++)
	{
    if (seqnums[i].value)
		{
			printf("Stream %lu (type 0x%02x, beginning seqnum = 0x%02x, ending seqnum = 0x%02x) had %u packets\n", i+1,
             seqnums[i].type, seqnums[i].beginning_seqnum, seqnums[i].value, seqnums[i].count);
 		}
	}
	
  putchar('\n');

  printf("Smallest packet found: %u bytes\n", min_packet_size);
  printf("Largest packet found: %u bytes\n\n", max_packet_size);

  printf("Wrote %lu chunks to %s\n", numchunks, argv[1]);
#endif

	net_close_and_remove(sock);
	asf_close_file(fd);

  return 0;
}
